from collections import defaultdict
from math import ceil, floor
from random import Random


def _meets_spec(project_input, section):
  if section['type'] == 'IMG_GROUP':
    return section['spec'].get('min', 0) <= len(project_input['data']['imgs']) <= section['spec']['max']

  if section['type'] == 'VIDEO':
    # TODO: you could also add validation over scale / crop stuff
    min_s = section['spec']['duration']['min_s']
    max_s = section['spec']['duration']['max_s']
    trim = project_input['data'].get('trim')
    if trim:
      duration = trim[1] - trim[0]
    else:
      duration = project_input['data']['duration']
    return min_s < duration < max_s

  # TODO: add validation over other types
  return True


def build_audio_choice_resource(audio_choice_input):
  return {
    'id': 'audio/%(input_id)s/%(version_id)i' % audio_choice_input,
    'type': 'AUDIO',
    'src': audio_choice_input['data']['src'],
    'metadata': {
      'duration': audio_choice_input['data']['duration']
    }
  }


def build_img_resource(img_input):
  return {
    'id': 'img/%(input_id)s/%(version_id)i' % img_input,
    'type': 'IMG',
    'src': img_input['data']['src'],
    'metadata': {
      'w': img_input['data']['size']['w'],
      'h': img_input['data']['size']['h']
    }
  }


def build_img_group_resources(img_group_input):
  for i, img in enumerate(img_group_input['data']['imgs']):
    yield {
      'id': 'img/%(input_id)s/%(version_id)i/%(i)i' % { **img_group_input, 'i': i },
      'type': 'IMG',
      'src': img['src'],
      'metadata': {
        'w': img['size']['w'],
        'h': img['size']['h']
      }
    }


def build_text_resource(text_input, section):
  # Note that we could copy a bunch of other properties into 'metadata' from the input, like the actual text, font,
  # color, etc. Presumably we'd only need to pass those through if there were a use case where a template wanted to do
  # something with those actual values.
  return {
    'id': 'text/%(input_id)s/%(version_id)i' % text_input,
    'type': 'IMG',
    'src': text_input['data']['src'],
    'metadata': {
      **section['spec']['bounds'],
      'text': text_input['data']['text'],
      'font': text_input['data'].get('font', section['spec']['font']),
      'color': text_input['data'].get('color', section['spec']['color']),
      'bold': text_input['data']['bold'],
      'italic': text_input['data']['italic'],
      'underline': text_input['data']['underline']
    }
  }


def build_video_resource(video_input):
  '''
  There is not a one-to-one correspondence between inputs of type 'VIDEO' and media resources of type 'VIDEO'. The
  inputs may have trim, crop, and rotate properties that need to be taken into account. This function accounts for
  those when building the media resource and includes metadata under the 'metadata' key.
  '''

  trim = video_input['data'].get('trim')
  if trim:
    t_offset = trim[0]
    duration = trim[1] - trim[0]
  else:
    t_offset = None
    duration = video_input['data']['duration']

  w = video_input['data']['size']['w']
  h = video_input['data']['size']['h']
  crop = video_input['data'].get('crop')
  if crop:
    w = w - w * crop['left'] - w * crop['right']
    h = h - h * crop['top'] - h * crop['bottom']

  return {
    'id': 'video/%(input_id)s/%(version_id)i' % video_input,
    'type': 'AUDIO_VIDEO' if video_input['data']['has_audio'] else 'VIDEO',
    'src': video_input['data']['src'],
    'crop': crop,
    't_offset': t_offset,
    'metadata': {
      'w': w,
      'h': h,
      'r': video_input['data']['r'],
      'scale': video_input['data'].get('scale'),
      'duration': duration
    }
  }


def _build_resources(project_input, section):
  section_type = section['type']
  if section_type == 'AUDIO_CHOICE':
    yield build_audio_choice_resource(project_input)
  elif section_type == 'IMG':
    yield build_img_resource(project_input)
  elif section_type == 'IMG_GROUP':
    yield from build_img_group_resources(project_input)
  elif section_type == 'TEXT':
    yield build_text_resource(project_input, section)
  elif section_type == 'VIDEO':
    yield build_video_resource(project_input)
  else:
    raise NotImplementedError()


class TemplateBuilder:

  def __init__(self, project, sections, add_resources=True):
    self._sections = {}
    for section in sections:
      self._sections[section['section_id']] = section

    # If you do use random numbers, it's important to only use a seeded random number generator so we end up with
    # consistent results:
    self.random = Random()
    self.random.seed(a=project.get('created_at'))

    # Automatically add resources for all the project sections - we might not need these, but if we never set any
    # automations for these resources, they'll never get used. Note that in Python 3, all dictionaries are ordered.
    self._resources = {}
    if add_resources:
      for section_id, project_section in project['sections'].items():
        for project_input in project_section['inputs']:
          section = self._sections.get(section_id)
          if section and _meets_spec(project_input, section):
            resources = _build_resources(project_input, self._sections[section_id])
            for resource in resources:
              resource['metadata']['section_id'] = section_id
              resource['metadata']['layer'] = len(self._resources)
              self._resources[resource['id']] = resource


  def get_resources(self, section_id=None):
    return [ x for x in self._resources.values() if (not section_id) or x.get('metadata', {}).get('section_id') == section_id ]


  def get_resource(self, section_id):
    resources = self.get_resources(section_id)
    if len(resources) > 1:
      raise Exception('This section has multiple resources - use get_resources instead of get_resource')
    return resources[0] if resources else None


  def add_resource(self, resource):
    if resource['id'] in self._resources:
      raise Exception("Resource with ID '%s' is already registered" % resource['id'])
    self._resources[resource['id']] = resource


  # Don't include resources whose total duration (as determined by their automations) does not meet this threshold.
  # This also effectively serves as our floating-point fudge factor when dealing with loop boundaries.
  _INCLUDE_DURATION_THRESHOLD = .01


  def _loop(self, resource, duration):
    # Yields multiple resources by splitting out the automations of this base resource such that the duration with
    # which we initialized this 'loop' will not be exceeded by the automations.
    # For example, given a looped resource with a duration of 4 and automations at times [ 10, 12, 20 ], we want
    # to create three separate resources with automations as follows:
    # [ 10, 12, 14 ]
    # [ 14, 18 ]
    # [ 18, 20 ]
    # The parameters associated with the entries for 14 and 18 should be linearly interpolated
    t0 = resource['automation'][0]['t']
    t1 = resource['automation'][-1]['t']

    # Note the floating point fudge factor to ensure we don't loop in the case where (t1 - t0) == duration
    n_loops = ceil((t1 - t0 - self._INCLUDE_DURATION_THRESHOLD) / duration)
    if n_loops <= 1:
      yield resource
      return

    # A structure for storing the automations of our resources
    loops = list(list() for _ in range(n_loops))

    # First, put our existing automations in the appropriate loop
    for a in resource['automation']:
      # If the boundaries happen to line up with the duration exactly, then our loop_i will be one too high for the
      # last automation. This problem should only happen with the final loop.
      loop_i = floor(abs(a['t'] - t0 - self._INCLUDE_DURATION_THRESHOLD) / duration)
      loops[loop_i].append(a)

    # Next make sure very loop has an automation that starts and ends at exactly the right time - note that the
    # first loop already starts at the right time and the last loop already ends at the right time
    for i in range(1, n_loops):
      new_a = { 't': t0 + i * duration }
      # We know the last loop must already have at least one automation
      last_a = loops[i - 1][-1]
      # We might actually have to look past a few loops until we find the next automation
      next_a = next(a for a in resource['automation'] if a['t'] > new_a['t'])
      pct = (new_a['t'] - last_a['t']) / (next_a['t'] - last_a['t'])
      for j in 'x', 'y', 'w', 'h', 'v':
        if (j in last_a) and (j in next_a):
          new_a[j] = last_a[j] + (next_a[j] - last_a[j]) * pct
      loops[i - 1].append(new_a)
      loops[i].insert(0, new_a)

    # Now we can use the loop automations to build our new resources
    for i, loop in enumerate(loops):
      looped_resource = { **resource, 'id': '%s:loop-%i' % (resource['id'], i), 'automation': loop }
      if (i != 0) and ('fade_in' in looped_resource):
        del looped_resource['fade_in']
      if (i != len(loops) - 1) and ('fade_out' in looped_resource):
        del looped_resource['fade_out']
      yield looped_resource


  def get_media(self):
    resources = []
    for resource in sorted(self._resources.values(), key=lambda r: r.get('metadata', {}).get('layer', 0)):
      if len(resource.get('automation', [])) < 2:
        # Not enough automations to include
        continue

      # Make sure all the automations are sorted by 't' and remove any duplicates
      resource['automation'].sort(key=lambda a: a['t'])
      unique = []
      for a in resource['automation']:
        if (not unique) or (unique[-1]['t'] != a['t']):
          unique.append(a)
      resource['automation'] = unique

      if resource['automation'][-1]['t'] - resource['automation'][0]['t'] < self._INCLUDE_DURATION_THRESHOLD:
        # Automations do not span a significant enough amount of time to include
        continue

      duration = resource.get('metadata', {}).get('duration')
      if duration:
        # Resources with a duration will loop automatically
        for looped_resource in self._loop(resource, duration):
          resources.append(looped_resource)
        continue

      resources.append(resource)


    return { 'w': 1920, 'h': 1080, 'media': resources }
