import base64
import io

from functools import lru_cache
from PIL import Image, ImageDraw, ImageFilter


def create_frame_image(w, h, thickness, color):
  '''
  Returns a PNG file containing the frame as defined with transparency in the middle
  '''
  img = Image.new('RGBA', (w, h), (0, 0, 0, 0))

  # First draw the shadow
  draw = ImageDraw.Draw(img)
  black = (0,0,0)
  draw.rectangle((2, 2, w - 2, thickness), fill=black)
  draw.rectangle((2, 2, thickness, h - 2), fill=black)
  draw.rectangle((w - 2, h - 2, w - 2 - thickness, 2), fill=black)
  draw.rectangle((w - 2, h - 2, 2, h - 2 - thickness), fill=black)
  for i in range(3):
    img = img.filter(ImageFilter.BLUR)

  # The frame itself
  draw = ImageDraw.Draw(img)
  draw.rectangle((2, 2, w - 2, thickness), fill=color)
  draw.rectangle((2, 2, thickness, h - 2), fill=color)
  draw.rectangle((w - 2, h - 2, w - 2 - thickness, 2), fill=color)
  draw.rectangle((w - 2, h - 2, 2, h - 2 - thickness), fill=color)

  in_mem_file = io.BytesIO()
  img.save(in_mem_file, 'PNG')
  in_mem_file.seek(0)
  return in_mem_file


@lru_cache
def _get_frame_url_cached(w, h, thickness, color):
  w, h, thickness = map(int, (w, h, thickness))
  file = create_frame_image(w, h, thickness, color)
  return 'data:image/png;base64,' + base64.b64encode(file.read()).decode('utf-8')


async def get_frame_url(w, h, thickness, color):
  return _get_frame_url_cached(w, h, thickness, color)


async def build_frame(resource, thickness, color, w=None, h=None):
  '''
  Builds a frame resource around the given visual media resource such that the frame will half overlap the edges of the
  visual media. This is a good mechanism for hiding artifacts caused by the lack of anti-aliasing for rotated media
  in rendered videos.

  Note: When adding this frame resource to the builder, make sure that its layer is greater than or equal to the layer
  of the input resource:
  >>> frame_resource = build_frame(resource, ...)
  >>> builder.add(frame_resource, layer=resource['_layer'])

  Params:
  - resource: the resource to frame, must be visual (IMG, GIF, or VIDEO) and have all of its automations in place
  - thickness: the thickness in pixels of the frame
  - color: an RGB tuple of three ints in the range [0,255]
  - w: the width of the frame, defaults to the width of the first automation
  - h: the height of the frame, defaults to the height of the first automation
  '''
  if not resource['automation']:
    raise Exception('Resource must have automation before a frame can be attached')
  w = w or resource['automation'][0]['w']
  h = h or resource['automation'][0]['h']
  metadata = {}
  if 'layer' in resource.get('metadata', {}):
    metadata['layer'] = resource['metadata']['layer'] + .1
  return {
    'id': resource['id'] + ':frame',
    'type': 'IMG',
    'src': await get_frame_url(w + thickness, h + thickness, thickness, color),
    'fade_in': resource.get('fade_in'),
    'fade_out': resource.get('fade_out'),
    'automation': [ { **a, 'w': a['w'] + thickness, 'h': a['h'] + thickness } for a in resource['automation'] ],
    'metadata': metadata
  }
