from .layout import get_scale


def slideshow_ken_burns(resources, t0, duration=5, crossfade=True, motion_x=40, motion_y=20, zoom=1.08):
  '''
  Attaches automations to the given IMG resources to make a slideshow using the ken-burns effect. Images that are not
  exactly 16x9 will have have some letterboxing that shows the background. Returns the total duration of the
  automations.

  Params:
  - resources: the IMG resources to put into the slideshow, should have the 'metadata' property set containing at
    least 'w' and 'h'
  - t0: the initial timestamp from which to begin the automations
  - duration: the amount of time each image is shown for. Note that there will be some overlap depending on if the
    following / preceeding image resources have fade_in / fade_out set.
  - crossfade: if set to True, will overlap successive images according to the fade_out time
  - motion_x: the amount the images should pan horizontally over the course of its automations
  - motion_y: the amount the images should pan vertically over the course of its automations
  - zoom: the zoom level at which to start (always ends at 1)
  '''

  t = t0
  for i, resource in enumerate(resources):
    size = resource['metadata']
    scale = get_scale(size['w'], size['h'], 1920 + 2 * motion_x, 1080 + 2 * motion_y, 'contain')
    w = scale * size['w']
    h = scale * size['h']
    # Start panning from top left
    c0 = 960 - motion_x, 540 - motion_y
    c1 = 960 + motion_x, 540 + motion_y
    if i % 2 != 0:
      # Odd numbered images will start panning from the bottom right
      c0, c1 = c1, c0
    resource['automation'] = [
      { 't': t, 'x': c0[0], 'y': c0[1], 'w': w * zoom, 'h': h * zoom, 'r': 0 },
      { 't': t + duration, 'x': c1[0], 'y': c1[1], 'w': w, 'h': h, 'r': 0 }
    ]
    if crossfade and (i != len(resources) - 1):
      t += duration - resource.get('fade_out', 0)
    else:
      t += duration
  return t - t0
