from math import ceil, sqrt


def _group(items, max_size):
  '''
  Returns a nested list of the items, grouped so that each child list does not exceed the 'max_size' and balanced so
  that the groups are as close to the same size as possible. If the groups do not come out evenly, then all the groups
  will have the same number of items save for the last group which will have fewer.

  For example:
  >>> group([ 1, 2, 3, 4, 5, 6 ], 2) => [ [ 1, 2 ], [ 3, 4 ], [ 5, 6 ] ]
  >>> group([ 1, 2, 3, 4, 5 ], 4) => [ [ 1, 2, 3 ], [ 4, 5 ] ]
  '''
  n_groups = ceil(len(items) / max_size)
  if not n_groups:
    return []
  max_group_size = ceil(len(items) / n_groups)
  groups = []
  group = []
  for i in items:
    group.append(i)
    if len(group) == max_group_size:
      groups.append(group)
      group = []
  if group:
    groups.append(group)
  return groups


def _scale_bounds_for_area(width, height, area):
  '''
  Returns a tuple of (width, height) adjusted so that it has the same aspect ratio as the input bounds, but its area is
  as given
  '''
  aspect_ratio = width / height
  h1 = sqrt(area / aspect_ratio)
  w1 = h1 * aspect_ratio
  return w1, h1


async def spread_offset_horizontally(resources, t0, max_group_size=4, top=40, right=40, bottom=40, left=40, center_y_offset_pct=0.1, rotation_offset=15, delay=0.25):
  '''
  Attaches automations to the given video resources to result in a random-ish zig-zag effect as seen during the
  smile-and-wave section of the holiday cheers template. Returns the total duration of the automations.

  Params:
  - resources: the VIDEO resources to lay out, should have the 'metadata' property set containing 'w', 'h', 'r', and
    'duration'
  - t0: the initial timestamp from which to begin the automations
  - max_group_size: the max number of videos to show at once
  - top: the top margin
  - right: the right margin
  - bottom: the bottom margin
  - left: the left margin
  - center_y_offset_pct: the percentage of height after applying the margins to which these videos should be offset
    from the center horizontal axis when zig-zagging. Set to 0 for a straight line. Set to .5 to maximize zig-zag
    effect.
  - rotation_offset: the angle in degrees to stagger each item
  - delay: the delay between adding successive items (in a single group)
  '''

  if not resources:
    return 0

  t = t0
  grouped = _group(resources, max_group_size)
  max_group_size = max(len(x) for x in grouped)
  col_width = (1920 - left - right) / max_group_size
  for group in grouped:
    max_duration = 0
    for i, resource in enumerate(group):
      metadata = resource['metadata']
      max_duration = max(max_duration, metadata['duration'])
      w, h = _scale_bounds_for_area(metadata['w'], metadata['h'], col_width * (1080 - top - bottom) * .5 + center_y_offset_pct)
      x_offset = (max_group_size - len(group)) * col_width / 2
      x = left + x_offset + i * col_width + col_width / 2
      y = top + (.5 - center_y_offset_pct if i % 2 == 0 else .5 + center_y_offset_pct) * (1080 - top - bottom)
      if len(group) == 1:
        r = metadata['r']
      else:
        r = metadata['r'] - rotation_offset + 2 * rotation_offset * i / (len(group) - 1)
      resource['automation'] = [
        { 't': t, 'x': x, 'y': y, 'w': w, 'h': h, 'r': r, 'v': 0 },
        { 't': t + metadata['duration'], 'x': x, 'y': y, 'w': w, 'h': h, 'r': r, 'v': 0 }
      ]
      t += delay
    t += max_duration
  return t - t0
